/*
Copyright (c) 2016 Microsoft Corporation. All rights reserved.
Released under Apache 2.0 license as described in the file LICENSE.

Author: Daniel Selsam
*/
#pragma once
namespace lean {
void initialize_inductive_compiler_module();
void finalize_inductive_compiler_module();
}
